#!/bin/bash
# Run this from top of Fvwm95 dist after purging all junk.

cat <<"EOF"
# Nonrecursive Makefile.am generated by makemakefileam.sh

SUFFIXES = .c .h .y .l
AUTOMAKE_OPTIONS = subdir-objects
AM_CPPFLAGS = -D_FORTIFY_SOURCE=2 -I$(top_srcdir)/include

# Warning: must synchronize with path setup in configure.ac and again in
# sample.fvwmrc/system.fvwm95rc.in.  These extended path variables don't
# transfer.
pixmapsdir = $(pkgdatadir)/pixmaps
iconsdir   = $(pkgdatadir)/mini-icons
pluginsdir = $(pkglibexecdir)/plugins
scriptsdir = $(pkgdatadir)/scripts

# Flex and bison
AM_YFLAGS = -d
# I tried *everything* else to make it NOT dist the built sources.
dist-hook:
	rm -f $(distdir)/modules/FvwmScript/bison.h $(distdir)/modules/FvwmScript/bison.c $(distdir)/modules/FvwmScript/flex.c

# Bins
bin_PROGRAMS = fvwm95 xpmroot

# Scripts that go in bin
dist_bin_SCRIPTS = utils/fvwmrc_convert utils/quantize_pixmaps

# Config file
nodist_sysconf_DATA = sample.fvwmrc/system.fvwm95rc

# Extra headers that every bin depends on
# include/FVWMconfig.h is omitted simply to stop it being disted.
BINHDRS = include/decorations.h include/shims.h include/fvwm/fvwmlib.h

EOF

echo "# Miscellany"
echo -n "dist_man1_MANS ="
find * -name \*.man -printf " %p"
echo ""
echo -n "dist_pixmaps_DATA ="
find pixmaps -name \*.x\* -printf " %p"
echo ""
echo -n "dist_icons_DATA ="
find mini-icons -name \*.x\* -printf " %p"
echo ""
echo -n "dist_html_DATA ="
find docs/html -name \*.html -printf " %p"
echo ""
# If suffixed SCRIPTS it makes them executable.
echo -n "dist_scripts_DATA ="
find modules/FvwmScript/Scripts -type f -a \! -name Makefile.in -printf " %p"
echo ""

echo ""

echo "# Libs used during build, not useful to install"
echo "noinst_LIBRARIES = libfvwm95.a libWidgets.a"
echo -n "libfvwm95_a_SOURCES = \$(BINHDRS) "
ls libs/*.[c,h] | xargs
echo -n "libWidgets_a_SOURCES = \$(BINHDRS) "
ls modules/FvwmScript/Widgets/*.[c,h] | xargs
echo ""

echo "# Bins"
echo "xpmroot_SOURCES = utils/xpmroot.c"
echo "xpmroot_LDADD = -lXpm -lX11"
echo -n "fvwm95_SOURCES = \$(BINHDRS) "
ls fvwm/*.[c,h] | xargs
echo "fvwm95_LDADD = libfvwm95.a -lXpm -lXext -lX11"
echo ""

echo "# Modules"
MODULES=`basename -a \`ls -d modules/Fvwm*\` | xargs`
echo "pkglibexec_PROGRAMS = ${MODULES} FvwmConsoleC"
for m in ${MODULES}; do
  echo -n "${m}_SOURCES = \$(BINHDRS) "

  # Exclude taskbar plugins (*Module.c)
  # Exclude FvwmConsoleC
  ls modules/${m}/*.[c,h,l,y] | fgrep -v Module.c | fgrep -v FvwmConsoleC.c | fgrep -v getline.c | xargs

  if [ "${m}" == "FvwmScript" ]; then
    echo "${m}_LDADD = libfvwm95.a libWidgets.a -lXpm -lXext -lX11"
    # The extra -I is needed for the built sources to survive distcheck.
    echo "${m}_CPPFLAGS = \$(AM_CPPFLAGS) -I\$(top_srcdir)/modules/FvwmScript"
  elif [ "${m}" == "FvwmTaskBar" ]; then
    echo "${m}_LDADD = libfvwm95.a -lXpm -lXext -lX11 -ldl"
    echo "${m}_LDFLAGS = -rdynamic"
  else
    echo "${m}_LDADD = libfvwm95.a -lXpm -lXext -lX11"
  fi
done
echo "FvwmConsoleC_SOURCES = \$(BINHDRS) modules/FvwmConsole/FvwmConsoleC.c modules/FvwmConsole/getline.c modules/FvwmConsole/FvwmConsole.h"
echo "FvwmConsoleC_LDADD = -lreadline -lncurses"
echo ""

echo "# Plugins - automake does not handle them well"
PLUGINS=`ls modules/FvwmTaskBar/*Module.c | xargs`
echo "plugins_DATA = ${PLUGINS//.c/.so}"
for p in ${PLUGINS}; do
  # Order matters: $< is the first prerequisite
  echo ${p%.c}.so: ${p} \$\(BINHDRS\)
  echo -e "\t\$(COMPILE) -rdynamic -shared -fPIC -o \$@ \$<"
done
echo ""

echo "# Everything that make dist misses"
echo -n "EXTRA_DIST = BUGS NEVERDO utils/makemakefileam.sh ${PLUGINS}"
find docs/modules -type f -printf " %p"
find docs/random -type f -printf " %p"
echo ""
echo ""

echo "# Everything that make mostlyclean misses"
echo "MOSTLYCLEANFILES = modules/FvwmScript/bison.h modules/FvwmScript/bison.c modules/FvwmScript/flex.c ${PLUGINS//.c/.so}"
